/* $Id: auth.c,v 1.8 2002/10/17 10:03:58 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <string.h>
#include <err.h>

#include <openssl/evp.h>

#include "auth.h"
#include "protocol.h"
#include "rpfcd.h"

/*
 * Username/password authentication. Loops thru rpfcd users file and
 * looks for a match on supplied username. If found, it calculates the
 * SHA-1 digest of the supplied password and checks against the digest
 * stored in the users file.
 *
 * Returns 0 on success, 1 on failure.
 */
int
rpfcd_auth_user(FILE *fp, char *username, char *password)
{
	char *buf, *lbuf, **ap, *args[2], digest[2 * SHA_DIGEST_LEN + 1];
	size_t len;
	int retval;

	lbuf = NULL;

	/*
	 * We need to rewind the file pointer, since it will puke if a
	 * previous authentication attempt failed.
	 */
	rewind(fp);
	
	while ((buf = fgetln(fp, &len))) {
		if (buf[len - 1] == '\n')
			buf[len - 1] = '\0';
		else {
			if ((lbuf = (char *)malloc(len + 1)) == NULL)
				err(1, NULL);
			memcpy(lbuf, buf, len);
			lbuf[len] = '\0';
			buf = lbuf;
		}

		if (lbuf != NULL) {
			free(lbuf);
			lbuf = NULL;
		}

		/* Separate the input string so we can look for a match on username */
		for (ap = args; ap < &args[2] && (*ap = strsep(&buf, ":")) != NULL;)
			if (**ap != '\0')
				ap++;

		*ap = NULL;

		if ((strcmp(username, args[0])) == 0) {
			memset(digest, '\0', sizeof(digest));
			retval = calc_sha1_digest(password, digest);

			if (retval < 1) {
				if ((strcmp(args[1], digest)) == 0)
					return (0);
				else
					return (1);
			}

			/* Don't look any further */
			break;
		}
	}

	return (1);
}

/*
 * Generic SHA-1 digest calculation. Uses OpenSSL EVP() wrappers instead
 * of low-level sha1() calls.
 *
 * Returns 0 on success, 1 on failure.
 */
int
calc_sha1_digest(char *input, char *digest)
{
	char *ptr;
	int md_len, i, retlen;
	unsigned char md_value[EVP_MAX_MD_SIZE];
	EVP_MD_CTX mdctx;
	const EVP_MD *md;

	if (input == NULL)
		return (1);

	OpenSSL_add_all_digests();
	md = EVP_sha1();

	EVP_MD_CTX_init(&mdctx);
	EVP_DigestInit_ex(&mdctx, md, NULL);
	EVP_DigestUpdate(&mdctx, input, strlen(input));
	EVP_DigestFinal_ex(&mdctx, md_value, &md_len);
	EVP_MD_CTX_cleanup(&mdctx);

	ptr = digest;
	retlen = 0;

	for(i = 0; i < md_len; i++)
		retlen = snprintf(ptr += retlen, 2 * SHA_DIGEST_LEN, "%02x", md_value[i]);

	return (0);
}

/*
 * Split the input into two separate strings, i.e. username and password.
 * Format must be like 'XXusernamepassword\r\n' where XX is length of username.
 */
int
parse_auth_request(char *input, char *username, char *password)
{
	char *ptr, ilen[2];
	size_t span;
	int len;

	ptr = input;

	memcpy(ilen, ptr, 2);
	len = (int)strtol(ilen, (char **)NULL, 10);

	if (len > MAX_USERNAME_LEN)
		return (1);

	memcpy(username, ptr += 2, len);
	span = strcspn(ptr += len, "\r\n");

	if (span > MAX_PASSWORD_LEN)
		return (1);

	memcpy(password, ptr, span);
	return (0);
}

/*
 * Open handle to RPFCD_USERS (tipically /etc/rpfcd/rpfcd.users) and return the
 * handle to the caller.
 */
FILE *
rpfcd_open_users_file(void)
{
	FILE *fp;
	
	fp = fopen(RPFCD_USERS, "r");
	return (fp == NULL) ? NULL : fp;
}
