/* $Id: rpfcd.c,v 1.13 2002/10/17 10:03:58 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <net/pfvar.h>
#include <netdb.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <err.h>

#include "auth.h"
#include "protocol.h"
#include "misc.h"
#include "pf_ioctl.h"
#include "rpfcd.h"
#include "ssl.h"

/*
 * Local rpfcd prototypes.
 */
static void show_rpfcd_version(void);
static void usage(void);
int	tcp_init(long);
void sig_child(int);

extern char *__progname;

/*
 * rpfcd is a program allowing remote control and management of OpenBSD pf (packet filter).
 * Communication is done using a custom protocol running on top of SSL.
 *
 * For a definition of protocol used see protocol.h.
 */
int
main(int argc, char *argv[])
{
	int ch, dev;
	long rpfcd_port;
	extern char *optarg;
	extern int optind;
	char *passphrase;
	char *ep;
	FILE *fp;

	int parent_socket = 0;
	int sock = 0;
	pid_t pid;
	SSL_CTX *ssl_ctx;
	SSL *ssl;
	BIO *sock_bio;

	int vflag, pflag, dflag;
	vflag = pflag = dflag = 0;

	while ((ch = getopt(argc, argv, "p:DV")) != -1)
		switch (ch) {
		case 'p':
			pflag = 1;
			rpfcd_port = strtol(optarg, &ep, 10);
			if (rpfcd_port <= 0 || *ep != '\0' || rpfcd_port > 65535)
				errx(1, "illegal port number");
			break;
		case 'D':
			dflag = 1;
			break;
		case 'V':
			vflag = 1;
			break;
		case '?':
		default:
			usage();
			/* Not reached */
		}
	
	argc -= optind;
	argv += optind;
 
 	if (vflag)
		show_rpfcd_version();

	if (!pflag)
		rpfcd_port = RPFCD_DEFAULT_PORT;

	parent_socket = tcp_init(rpfcd_port);
	ssl_ctx = ssl_ctx_init(RPFCD_CERT, passphrase);

	/* We need to handle child zombies */
	signal(SIGCHLD, sig_child);

	for (;;) {
		if (dflag)
			if (daemon(0, 0) < 0)
				errx(1, "daemon() failed: %.200s", strerror(errno));

		if ((sock = accept(parent_socket, 0, 0)) < 0)
			errx(1, "error accepting connection: %s", strerror(errno));

		if ((pid = fork()))
			close(sock);
		else {
			/* Open /dev/pf handle while still UID 0 */
			dev = pf_open_device();
			if (dev <= 0)
				errx(1, "unable to open %s: %s", PF_DEVICE, strerror(errno));

			/* Open RPFCD_USERS while still UID 0 */
			fp = rpfcd_open_users_file();
			if (fp == NULL)
				errx(1, "unable to open %s: %s", RPFCD_USERS, strerror(errno));

			if ((chroot(RPFCD_CHROOT)) != 0)
				errx(1, "unable to chroot to %s: %s", RPFCD_CHROOT, strerror(errno));

			/* Now drop privileges */
			seteuid(RPFCD_UID);
			setuid(RPFCD_UID);

			sock_bio = BIO_new_socket(sock, BIO_NOCLOSE);
			ssl = SSL_new(ssl_ctx);
			SSL_set_bio(ssl, sock_bio, sock_bio);
			
			if ((SSL_accept(ssl)) <= 0)
				errx(1, "SSL accept error");

			ssl_handle_request(ssl, sock, dev, fp);

			if ((pf_close_device(dev)) < 0)
				warn("unable to close %s", PF_DEVICE);

			exit(EXIT_SUCCESS);
		}
	}

	ssl_ctx_terminate(ssl_ctx);
	exit(EXIT_SUCCESS);
}

/*
 * Initialize a new TCP session.
 */
int
tcp_init(long port)
{
	int sock;
	struct sockaddr_in sock_in;
	int dummy = 1;

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		errx(1, "error creating socket");

	memset(&sock_in, 0, sizeof(sock_in));
	sock_in.sin_addr.s_addr = htonl(INADDR_ANY);
	sock_in.sin_family = AF_INET;
	sock_in.sin_port = htons(port);
	
	if ((setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof(dummy))) != 0)
		errx(1, "error setting socket options");

	if ((bind(sock, (struct sockaddr *)&sock_in, sizeof(sock_in))) < 0)
		errx(1, "error binding to %ld", port);

	if ((listen(sock, 5)) != 0)
		errx(1, "error setting client limit");

	return sock;
}

/*
 * Handle those pesky children zombies.
 */
void
sig_child(int signo)
{
	pid_t pid;
	int stat;

	while ((pid = waitpid(-1, &stat, WNOHANG)) > 0)

	return;
}
	
/*
 * Show rpfcd version, including protocol and OpenSSL version.
 */
static void
show_rpfcd_version(void)
{
	(void)fprintf(stderr, "%s, protocol %d.%d, OpenSSL 0x%8.8lx\n",
		RPFCD_VERSION, RPFCP_MAJOR, RPFCP_MINOR, SSLeay());
	exit(EXIT_FAILURE);
}

/*
 * Show usage information.
 */
static void
usage(void)
{
	(void)fprintf(stderr, "Usage: %s [options]\n", __progname);
	(void)fprintf(stderr, "Options:\n");
	(void)fprintf(stderr, "  -D          Run as daemon (no debug info)\n");
	(void)fprintf(stderr, "  -p port     Bind to port (default 2620)\n");
	(void)fprintf(stderr, "  -V          Show version information\n");
	exit(EXIT_FAILURE);
}
