/* $Id: ssl.c,v 1.22 2002/10/17 19:47:46 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <net/if.h>
#include <net/pfvar.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <err.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

#include "auth.h"
#include "misc.h"
#include "protocol.h"
#include "pf_ioctl.h"
#include "ssl.h"

BIO *bio_err = 0;
char *pass;

/*
 * Handle incoming SSL requests from clients.
 */
int
ssl_handle_request(SSL *ssl, int sock, int dev, FILE *fp)
{
	char request[RPFCP_MAX_MESSAGE_SIZE + 1];
	char response[RPFCP_MAX_MESSAGE_SIZE + 1];
	char username[MAX_USERNAME_LEN + 1], password[MAX_PASSWORD_LEN + 1];
	int retval, len, rules_count;
	int multi_line = 0;
	int authenticated = 0;

	struct rpfcp_request r_req;
	struct rpfcp_response r_resp;

	BIO *io, *ssl_bio;

	io = BIO_new(BIO_f_buffer());
	ssl_bio = BIO_new(BIO_f_ssl());
	BIO_set_ssl(ssl_bio, ssl, BIO_CLOSE);
	BIO_push(io, ssl_bio);

	for (;;) {
		for (;;) {
			/* We reuse our buffers, zero them after each iteration */
			memset(request, '\0', sizeof(request));
			memset(response, '\0', sizeof(response));
			memset(&r_req, '\0', sizeof(struct rpfcp_request));
			memset(&r_resp, '\0', sizeof(struct rpfcp_response));

			retval = BIO_gets(io, request, RPFCP_MAX_MESSAGE_SIZE);

			switch (SSL_get_error(ssl, retval)) {
			case SSL_ERROR_NONE:
				len = retval;
				break;
			default:
				errx(1, "error reading ssl data");
			}
			break;
		}

		/* We have the client request, validate and proceed if ok. */
		if ((parse_RPFCP_request(&r_req, request)) != 0)
			r_resp.code = RPFCP_VIOLATION;	
		else {
			if (r_req.code == RPFCP_DISCONNECT)
				goto terminate;

			if (!authenticated) {
				if (r_req.code == RPFCP_AUTH) {
					memset(username, '\0', sizeof(username));
					memset(password, '\0', sizeof(password));

					if ((parse_auth_request(r_req.msg, username, password)) != 0) {
						retval = RPFCP_VIOLATION;
						goto skip_switch;
					}

					if ((rpfcd_auth_user(fp, username, password)) != 0)
						retval = RPFCP_AUTH_FAILED;
					else {
						retval = RPFCP_NO_ERROR;
						authenticated = 1;

						/* Zero memory containing sensitive data. */
						memset(username, '\0', sizeof(username));
						memset(password, '\0', sizeof(password));

						/*
						 * Close opened handle to RPFCD_USERS. We don't need
						 * it any more since the user is authenticated.
						 */
						fclose(fp);
					}
				}
				else
					retval = RPFCP_AUTH_REQUIRED;

				goto skip_switch;
			}

			switch (r_req.code) {
			case RPFCP_REQ_PF_UPTIME:
				retval = pf_get_stats(dev, &r_resp, RPFCP_REQ_PF_UPTIME);
				break;
			case RPFCP_REQ_PF_STATES:
				retval = pf_get_stats(dev, &r_resp, RPFCP_REQ_PF_STATES);
				break;
			case RPFCP_REQ_PF_LIMITS:
				retval = pf_get_limits(dev, &r_resp);
				break;
			case RPFCP_REQ_PF_COUNTERS:
				retval = pf_get_stats(dev, &r_resp, RPFCP_REQ_PF_COUNTERS);
				break;
			case RPFCP_REQ_PF_TIMEOUTS:
				retval = pf_get_timeouts(dev, &r_resp);
				break;
			case RPFCP_REQ_PF_STATUS:
				retval = pf_get_status(dev, &r_resp);
				break;
			case RPFCP_REQ_PF_FILTER:
				retval = pf_get_number_of_filter_rules(dev, &rules_count);
				multi_line = 1;
				break;
			default:
				break;
			}

skip_switch:

			switch (retval) {
			case PF_ACCESS_DENIED:
				r_resp.code = RPFCP_PF_ACCESS_DENIED;
				break;
			case PF_QUERY_FAILED:
				r_resp.code = RPFCP_PF_QUERY_FAILED;
				break;
			case PF_DISABLED:
				r_resp.code = RPFCP_PF_DISABLED;
				break;
			case RPFCP_AUTH_REQUIRED:
				r_resp.code = RPFCP_AUTH_REQUIRED;
				break;
			case RPFCP_AUTH_FAILED:
				r_resp.code = RPFCP_AUTH_FAILED;
				break;
			case RPFCP_VIOLATION:
				r_resp.code = RPFCP_VIOLATION;
				break;
			default:
				r_resp.code = RPFCP_NO_ERROR;
				break;
			}
		}

		if (multi_line) {

			/* Send MULTILINE_BEGIN to indicate start of message. */
			r_resp.code = RPFCP_MULTILINE_BEGIN;
			r_resp.len = strlen(r_resp.msg) + 2;
			make_RPFCP_response(&r_resp, response);
			
			if ((retval = BIO_puts(io, response)) <= 0)
				errx(1, "error writing back to client");

			r_resp.code = RPFCP_NO_ERROR;

			if (r_req.code == RPFCP_REQ_PF_FILTER) {
				int i;

				for (i = 0; i < rules_count; ++i) {
					retval = pf_get_filter_rules(dev, i, &r_resp);
					r_resp.len = strlen(r_resp.msg) + 2;
					make_RPFCP_response(&r_resp, response);

					if ((retval = BIO_puts(io, response)) <= 0)
						errx(1, "error writing back to client");
				}
			}

			/* Send MULTILINE_END to indicate end of message. */
			r_resp.code = RPFCP_MULTILINE_END;
			r_resp.msg[0] = '\0';
            r_resp.len = strlen(r_resp.msg) + 2;
            make_RPFCP_response(&r_resp, response);
            
            if ((retval = BIO_puts(io, response)) <= 0)
                errx(1, "error writing back to client");

			if ((retval = BIO_flush(io)) < 0)
				errx(1, "error flushing BIO");

			/* Reset to avoid further message processing. */
			multi_line = 0;
		} else { 
			r_resp.len = strlen(r_resp.msg) + 2;
			make_RPFCP_response(&r_resp, response);

			if ((retval = BIO_puts(io, response)) <= 0)
				errx(1, "error writing back to client");

		if ((retval = BIO_flush(io)) < 0)
			errx(1, "error flushing BIO");
		}
	}

terminate:

    retval = SSL_shutdown(ssl);

    if (!retval) {
        shutdown(sock, 1);
        retval = SSL_shutdown(ssl);
    }

    SSL_free(ssl);
    close(sock);

    return (0);
}

/*
 * Password handling routine.
 */
int
pem_passwd_cb(char *buf, int size, int rwflag, void *userdata)
{
	if (size < strlen(pass) + 1)
		return 0;
	
	strlcpy(buf, pass, MAX_PASSPHRASE_LEN);
	return strlen(buf);
}

/*
 * SSL context initialization.
 */
SSL_CTX
*ssl_ctx_init(char *certificate, char *passphrase)
{
	SSL_METHOD *ssl_method;
	SSL_CTX *ssl_ctx;

	OpenSSL_add_ssl_algorithms();
	SSL_load_error_strings();

	bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);

	ssl_method = SSLv23_server_method();
	ssl_ctx = SSL_CTX_new(ssl_method);

	if ((SSL_CTX_set_cipher_list(ssl_ctx, RPFCD_SSL_CIPHERS)) != 1)
		errx(1, "unable to set any of the selected ciphers");

	if ((SSL_CTX_use_certificate_chain_file(ssl_ctx, certificate)) != 1)
		errx(1, "error loading certificate file");

	pass = passphrase;
	SSL_CTX_set_default_passwd_cb(ssl_ctx, pem_passwd_cb);

	/* Is the private key encrypted? */
	if ((SSL_CTX_use_PrivateKey_file(ssl_ctx, certificate, SSL_FILETYPE_PEM)) != 1) {

		/* We need a passphrase to decrypt the private key */
		if ((pass = read_passphrase("Enter passphrase:")) == NULL)
			errx(1, "error reading passphrase");

		/* Try again with the supplied passphrase */
		SSL_CTX_set_default_passwd_cb(ssl_ctx, pem_passwd_cb);
		if ((SSL_CTX_use_PrivateKey_file(ssl_ctx, certificate, SSL_FILETYPE_PEM)) != 1)
			errx(1, "error loading private key (%s)", certificate);
	}
		
	return (ssl_ctx);
}

/*
 * SSL context termination.
 */
void
ssl_ctx_terminate(SSL_CTX *ssl_ctx)
{
	SSL_CTX_free(ssl_ctx);
}
